@@ -1,12 +1,16 @@
+import shlex
+from subprocess import PIPE
+import subprocess
 from .definitions import (
     DocumentType, URL_FILENAME_MAX_LEN, Verbosity, InteractiveResult,
     SeleniumDownloadStrategy, ScrFetchError, SeleniumVariant, DEFAULT_CWF
 )
 from .input_sequences import (
     YES_INDICATING_STRINGS, NO_INDICATING_STRINGS, EDIT_INDICATING_STRINGS,
-    CHAIN_SKIP_INDICATING_STRINGS, DOC_SKIP_INDICATING_STRINGS
+    CHAIN_SKIP_INDICATING_STRINGS, DOC_SKIP_INDICATING_STRINGS,
+    INSPECT_INDICATING_STRINGS
 )
-from typing import Optional, BinaryIO, Union, cast, Iterator
+from typing import Any, Optional, BinaryIO, Union, cast, Iterator
 import os
 import urllib
 import time
@@ -17,7 +21,8 @@ from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
 import binascii
 import requests
 from . import (
-    progress_report, content_match, scr, utils, scr_context, selenium_setup
+    progress_report, content_match, scr, utils, scr_context, selenium_setup, locator,
+    document
 )
 from collections import OrderedDict
 import threading
@@ -38,14 +43,60 @@ class ContentFormat(Enum):
     UNNEEDED = 5,
 
 
+class ByteBuffer:
+    buffer: bytearray
+
+    def __init__(self) -> None:
+        self.buffer = bytearray()
+
+    def write(self, buffer: bytes, stderr: bool = False) -> int:
+        self.buffer.extend(buffer)
+        return len(buffer)
+
+    def write_str(self, buffer: str,  stderr: bool = False) -> int:
+        return self.write(buffer.encode("utf-8", errors="surrogateescape"), stderr)
+
+    def flush(self) -> None:
+        pass
+
+    def close(self) -> None:
+        pass
+
+    def to_bytes(self) -> bytearray:
+        return self.buffer
+
+
+class PrintOutputStream:
+    pom: 'PrintOutputManager'
+    id: int
+
+    def __init__(self, pom: 'PrintOutputManager') -> None:
+        self.pom = pom
+        self.id = pom.request_print_access()
+
+    def write(self, buffer: bytes, stderr: bool = False) -> int:
+        self.pom.print(self.id, buffer, stderr)
+        return len(buffer)
+
+    def write_str(self, buffer: str,  stderr: bool = False) -> int:
+        return self.write(buffer.encode("utf-8", errors="surrogateescape"), stderr)
+
+    def flush(self) -> None:
+        self.pom.flush(self.id)
+
+    def close(self) -> None:
+        self.pom.declare_done(self.id)
+
+
 class PrintOutputManager:
-    printing_buffers: OrderedDict[int, list[bytes]]
+    printing_buffers: OrderedDict[int, list[tuple[bytes, bool]]]
     finished_queues: set[int]
     lock: threading.Lock
     size_blocked: threading.Condition
     size_limit: int
     dl_ids: int = 0
     active_id: int = 0
+    active_id_stderr: bool = False
     main_thread_id: Optional[int] = None
 
     def __init__(self, max_buffer_size: int = DEFAULT_MAX_PRINT_BUFFER_CAPACITY) -> None:
@@ -57,6 +108,7 @@ class PrintOutputManager:
 
     def reset(self) -> None:
         self.active_id = 0
+        self.active_id_stderr = False
         self.dl_ids = 0
         self.main_thread_id = self.request_print_access()
 
@@ -65,7 +117,7 @@ class PrintOutputManager:
             self.declare_done(self.main_thread_id)
             self.main_thread_id = None
 
-    def print(self, id: int, buffer: bytes) -> None:
+    def print(self, id: int, buffer: bytes, stderr: bool) -> None:
         is_active = False
         with self.lock:
             while True:
@@ -73,19 +125,21 @@ class PrintOutputManager:
                     is_active = True
                     stored_buffers = self.printing_buffers.pop(id, [])
                     self.size_limit += sum(
-                        map(lambda b: len(b), stored_buffers)
+                        map(lambda b: len(b[0]), stored_buffers)
                     )
                     break
                 elif self.size_limit > len(buffer):
                     self.size_limit -= len(buffer)
-                    self.printing_buffers[id].append(buffer)
+                    self.printing_buffers[id].append((buffer, stderr))
                     break
                 self.size_blocked.wait()
         if is_active:
-            for b in stored_buffers:
-                sys.stdout.buffer.write(b)
-            sys.stdout.buffer.write(buffer)
-            if(stored_buffers):
+            for (b, stderr) in [*stored_buffers, (buffer, stderr)]:
+                if stderr:
+                    sys.stderr.buffer.write(b)
+                else:
+                    sys.stdout.buffer.write(b)
+            if len(stored_buffers) != 0:
                 self.size_blocked.notifyAll()
 
     def request_print_access(self) -> int:
@@ -106,12 +160,15 @@ class PrintOutputManager:
 
     def declare_done(self, id: int) -> None:
         new_active_id = None
-        buffers_to_print: list[list[bytes]] = []
+        buffers_to_print: list[list[tuple[bytes, bool]]] = []
         with self.lock:
             if self.active_id != id:
                 self.finished_queues.add(id)
                 return
 
+            if id in self.printing_buffers:
+                buffers_to_print.append(self.printing_buffers.pop(id))
+
             new_active_id = self.active_id + 1
             while new_active_id in self.finished_queues:
                 self.finished_queues.remove(new_active_id)
@@ -119,10 +176,14 @@ class PrintOutputManager:
                     self.printing_buffers.pop(new_active_id)
                 )
                 new_active_id += 1
+
         while True:
             for bl in buffers_to_print:
-                for b in bl:
-                    sys.stdout.buffer.write(b)
+                for (b, stderr) in bl:
+                    if stderr:
+                        sys.stderr.buffer.write(b)
+                    else:
+                        sys.stdout.buffer.write(b)
             # after we printed and reacquire the lock, the job
             # that we want to give the active_id token to
             # might have finished already, in which case we have to print him too
@@ -150,25 +211,6 @@ class PrintOutputManager:
         sys.stdout.flush()
 
 
-class PrintOutputStream:
-    pom: PrintOutputManager
-    id: int
-
-    def __init__(self, pom: PrintOutputManager) -> None:
-        self.pom = pom
-        self.id = pom.request_print_access()
-
-    def write(self, buffer: bytes) -> int:
-        self.pom.print(self.id, buffer)
-        return len(buffer)
-
-    def flush(self) -> None:
-        self.pom.flush(self.id)
-
-    def close(self) -> None:
-        self.pom.declare_done(self.id)
-
-
 class MinimalInputStream(ABC):
     @abstractmethod
     def read(self, size: Optional[int]) -> bytes:
@@ -247,12 +289,14 @@ class DownloadJob:
     temp_file_path: Optional[str] = None
     multipass_file: Optional[BinaryIO] = None
     print_stream: Optional[PrintOutputStream] = None
+    shell_cmd_stream: Optional[PrintOutputStream] = None
     content_stream: Union[BinaryIO, MinimalInputStream, None] = None
     content: Union[str, bytes, BinaryIO, MinimalInputStream, None] = None
     content_format: Optional[ContentFormat] = None
-    filename: Optional[str] = None
     status_report: Optional['progress_report.DownloadStatusReport'] = None
-
+    shell_proc: Optional[subprocess.Popen[Any]] = None
+    shell_output_handlers: list[concurrent.futures.Future[None]] = []
+    forward_buffer: Optional['ByteBuffer'] = None
     cm: 'content_match.ContentMatch'
     save_path: Optional[str] = None
     context: str
@@ -265,12 +309,21 @@ class DownloadJob:
         )
         self.output_formatters = []
 
+    def log(self, verbosity: Verbosity, msg: str) -> None:
+        if scr.check_log_message_needed(self.cm.mc.ctx, verbosity):
+            if self.status_report is not None:
+                self.status_report.error = msg
+            else:
+                scr.log(self.cm.mc.ctx, verbosity, msg)
+
     def requires_download(self) -> bool:
         return self.cm.mc.need_content and not self.cm.mc.content_raw
 
-    def setup_print_stream(self, pom: 'PrintOutputManager') -> None:
+    def request_print_streams(self, pom: 'PrintOutputManager') -> None:
         if self.cm.mc.content_print_format is not None:
             self.print_stream = PrintOutputStream(pom)
+        if self.cm.mc.content_shell_command_format is not None and self.cm.mc.content_shell_command_print_output:
+            self.shell_cmd_stream = PrintOutputStream(pom)
 
     def request_status_report(self, download_manager: 'DownloadManager') -> None:
         self.status_report = progress_report.DownloadStatusReport(
@@ -278,26 +331,89 @@ class DownloadJob:
         )
 
     def gen_fallback_filename(self, dont_use_url: bool = False) -> bool:
-        if self.filename is not None or not self.cm.mc.need_filename:
+        if self.cm.filename is not None or not self.cm.mc.need_filename:
             return True
         if not dont_use_url:
             path = cast(urllib.parse.ParseResult, self.cm.url_parsed).path
-            self.filename = scr.sanitize_filename(urllib.parse.unquote(path))
-            if self.filename is not None and len(self.filename) < URL_FILENAME_MAX_LEN:
+            self.cm.filename = scr.sanitize_filename(urllib.parse.unquote(path))
+            if self.cm.filename is not None and len(self.cm.filename) < URL_FILENAME_MAX_LEN:
                 return True
         try:
-            self.filename = scr.gen_final_content_format(
-                cast(str, self.cm.mc.filename_default_format), self.cm, None
+            self.cm.filename = scr.gen_final_content_format(
+                cast(str, self.cm.mc.filename_default_format), self.cm
             ).decode("utf-8", errors="surrogateescape")
             return True
         except UnicodeDecodeError:
-            scr.log(
-                self.cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{self.cm.doc.path}{scr.get_ci_di_context(self.cm)}: "
                 + "generated default filename not valid utf-8"
             )
             return False
 
+    def handle_label_match(self) -> InteractiveResult:
+        cm = self.cm
+        if not cm.mc.need_label or (cm.llm is not None and cm.llm.fres is not None):
+            # this was already done during for interactive filename determination
+            return InteractiveResult.ACCEPT
+
+        if cm.mc.need_filename_for_interaction:
+            if not self.fetch_content():
+                return InteractiveResult.ERROR
+
+        if cm.llm is None:
+            if cm.mc.need_label:
+                cm.llm = locator.LocatorMatch()
+                cm.llm.fres = cast(str, cm.mc.label_default_format).format(
+                    **scr.content_match_build_format_args(cm)
+                )
+                cm.llm.result = cm.llm.fres
+        else:
+            cm.mc.loc_label.apply_format_for_content_match(cm, cm.llm)
+
+        if cm.mc.loc_label.interactive:
+            di_ci_context = scr.get_ci_di_context(cm)
+            content_type = scr.get_content_type_label(cm)
+            assert cm.llm is not None
+            while True:
+                if not cm.mc.is_valid_label(cm.llm.result):
+                    self.log(
+                        Verbosity.WARN,
+                        f'"{cm.doc.path}": labels cannot contain a slash ("{cm.llm.result}")'
+                    )
+                else:
+                    prompt_options = [
+                        (InteractiveResult.ACCEPT, YES_INDICATING_STRINGS),
+                        (InteractiveResult.REJECT, NO_INDICATING_STRINGS),
+                        (InteractiveResult.EDIT, DOC_SKIP_INDICATING_STRINGS),
+                        (InteractiveResult.SKIP_CHAIN, CHAIN_SKIP_INDICATING_STRINGS),
+                        (InteractiveResult.SKIP_DOC, DOC_SKIP_INDICATING_STRINGS)
+                    ]
+                    if cm.mc.content_raw:
+                        prompt_options.append(
+                            (InteractiveResult.INSPECT, INSPECT_INDICATING_STRINGS))
+                        inspect_opt_str = "/inspect"
+                        prompt_msg = f'"{cm.doc.path}"{di_ci_context}: accept content label "{cm.llm.result}"'
+                    else:
+                        inspect_opt_str = ""
+                        prompt_msg = f'"{cm.doc.path}": {content_type} {cm.clm.result}{di_ci_context}: accept content label "{cm.llm.result}"'
+
+                    res = scr.prompt(
+                        f'{prompt_msg} [Yes/no/edit/{inspect_opt_str}/chainskip/docskip]? ',
+                        prompt_options,
+                        InteractiveResult.ACCEPT
+                    )
+                    if res == InteractiveResult.ACCEPT:
+                        break
+                    if res == InteractiveResult.INSPECT:
+                        print(
+                            f'"{cm.doc.path}": {content_type} for "{cm.llm.result}":\n' + cm.clm.result)
+                        continue
+                    if res != InteractiveResult.EDIT:
+                        return res
+                cm.llm.result = input("enter new label: ")
+        return InteractiveResult.ACCEPT
+
     def handle_save_path(self) -> 'InteractiveResult':
         if self.save_path is not None:
             # this was already done during for interactive filename determination
@@ -306,31 +422,31 @@ class DownloadJob:
         if not cm.mc.content_save_format:
             return InteractiveResult.ACCEPT
         if cm.llm and not cm.mc.is_valid_label(cm.llm.result):
-            scr.log(cm.mc.ctx, Verbosity.WARN,
-                    f"matched label '{cm.llm.result}' would contain a slash, skipping this content from: {cm.doc.path}"
-                    )
+            self.log(
+                Verbosity.WARN,
+                f"matched label '{cm.llm.result}' would contain a slash, skipping this content from: {cm.doc.path}"
+            )
             save_path = None
-        if cm.mc.need_filename:
+        if cm.mc.need_filename_for_interaction:
             if not self.fetch_content():
                 return InteractiveResult.ERROR
-        save_path_bytes = scr.gen_final_content_format(
-            cm.mc.content_save_format, cm, self.filename
-        )
+        save_path_bytes = scr.gen_final_content_format(cm.mc.content_save_format, cm)
         try:
             save_path = save_path_bytes.decode(
                 "utf-8", errors="surrogateescape"
             )
         except UnicodeDecodeError:
-            scr.log(
-                cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{cm.doc.path}{scr.get_ci_di_context(cm)}: generated save path is not valid utf-8"
             )
             save_path = None
         while True:
             if save_path and not os.path.exists(os.path.dirname(os.path.abspath(save_path))):
-                scr.log(cm.mc.ctx, Verbosity.ERROR,
-                        f"{cm.doc.path}{scr.get_ci_di_context(cm)}: directory of generated save path does not exist"
-                        )
+                self.log(
+                    Verbosity.ERROR,
+                    f"{cm.doc.path}{scr.get_ci_di_context(cm)}: directory of generated save path does not exist"
+                )
                 save_path = None
             if not save_path and not cm.mc.save_path_interactive:
                 return InteractiveResult.ERROR
@@ -359,6 +475,15 @@ class DownloadJob:
         self.save_path = save_path
         return InteractiveResult.ACCEPT
 
+    def handle_user_interaction(self) -> 'InteractiveResult':
+        res = InteractiveResult.ACCEPT
+
+        res = self.handle_label_match()
+        # TODO: handle skip chain/doc properly
+        if res.accepted():
+            return self.handle_save_path()
+        return res
+
     def selenium_download_from_local_file(self) -> bool:
         path = self.cm.clm.result
         self.content_format = ContentFormat.FILE
@@ -369,7 +494,7 @@ class DownloadJob:
                     offs += 1
             path = path[offs:]
         self.content = path
-        self.filename = os.path.basename(path)
+        self.cm.filename = os.path.basename(path)
         return True
 
     def selenium_download_external(self) -> bool:
@@ -379,7 +504,6 @@ class DownloadJob:
             proxies = {
                 "http": f"socks5h://localhost:{tbdriver.socks_port}",
                 "https": f"socks5h://localhost:{tbdriver.socks_port}",
-                "data": None
             }
         try:
             try:
@@ -391,7 +515,7 @@ class DownloadJob:
                 )
                 self.content = ResponseStreamWrapper(req)
                 self.content_format = ContentFormat.STREAM
-                self.filename = scr.request_try_get_filename(req)
+                self.cm.filename = scr.request_try_get_filename(req)
                 if self.status_report:
                     self.status_report.expected_size = (
                         scr.request_try_get_filesize(req)
@@ -400,8 +524,8 @@ class DownloadJob:
             except requests.exceptions.RequestException as ex:
                 raise scr.request_exception_to_scr_fetch_error(ex)
         except ScrFetchError as ex:
-            scr.log(
-                self.cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{utils.truncate(self.cm.doc.path)}{scr.get_ci_di_context(self.cm)}: "
                 + f"failed to download '{utils.truncate(self.cm.clm.result)}': {str(ex)}"
             )
@@ -414,8 +538,8 @@ class DownloadJob:
         doc_url = urllib.parse.urlparse(doc_url_str)
 
         if doc_url.netloc != cast(urllib.parse.ParseResult, self.cm.url_parsed).netloc:
-            scr.log(
-                self.cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{self.cm.clm.result}{scr.get_ci_di_context(self.cm)}: "
                 + "failed to download: seldl=internal does not work across origins"
             )
@@ -439,8 +563,8 @@ class DownloadJob:
             if selenium_setup.selenium_has_died(self.cm.mc.ctx):
                 selenium_setup.report_selenium_died(self.cm.mc.ctx)
             else:
-                scr.log(
-                    self.cm.mc.ctx, Verbosity.ERROR,
+                self.log(
+                    Verbosity.ERROR,
                     f"{self.cm.clm.result}{scr.get_ci_di_context(self.cm)}: "
                     + f"selenium download failed: {str(ex)}"
                 )
@@ -514,8 +638,8 @@ class DownloadJob:
             cors_warn = ""
             if urllib.parse.urlparse(doc_url).netloc != urllib.parse.urlparse(self.cm.clm.result).netloc:
                 cors_warn = " (potential CORS issue)"
-            scr.log(
-                self.cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{utils.truncate(self.cm.doc.path)}{scr.get_ci_di_context(self.cm)}: "
                 + f"selenium download of '{self.cm.clm.result}' failed{cors_warn}: {err}"
             )
@@ -523,7 +647,7 @@ class DownloadJob:
         self.content = binascii.a2b_base64(res["ok"])
         if self.status_report:
             self.status_report.expected_size = len(self.content)
-        self.filename = scr.try_get_filename_from_content_disposition(
+        self.cm.filename = scr.try_get_filename_from_content_disposition(
             res.get("content_disposition", "")
         )
         self.content_format = ContentFormat.BYTES
@@ -603,7 +727,7 @@ class DownloadJob:
                                 stream=True
                             )
                             self.content = ResponseStreamWrapper(res)
-                            self.filename = scr.request_try_get_filename(res)
+                            self.cm.filename = scr.request_try_get_filename(res)
                             if self.status_report:
                                 self.status_report.expected_size = (
                                     scr.request_try_get_filesize(res)
@@ -613,8 +737,13 @@ class DownloadJob:
                                 return False
                         except requests.exceptions.RequestException as ex:
                             fe = scr.request_exception_to_scr_fetch_error(ex)
-                            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
-                                    f"{self.context}: failed to download '{utils.truncate(self.cm.clm.result)}': {str(fe)}")
+                            if self.status_report:
+                                self.status_report.error = str(fe)
+                            else:
+                                self.log(
+                                    Verbosity.ERROR,
+                                    f"{self.context}: failed to download '{utils.truncate(self.cm.clm.result)}': {str(fe)}"
+                                )
                             return False
         return True
 
@@ -636,19 +765,21 @@ class DownloadJob:
             if use_as_multipass:
                 self.multipass_file = save_file
         except FileExistsError:
-            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
-                    f"{self.context}: file already exists: {self.save_path}")
+            self.log(
+                Verbosity.ERROR,
+                f"{self.context}: file already exists: {self.save_path}"
+            )
             return False
         except OSError as ex:
-            scr.log(
-                self.cm.mc.ctx, Verbosity.ERROR,
+            self.log(
+                Verbosity.ERROR,
                 f"{self.context}: failed to write to file '{self.save_path}': {str(ex)}"
             )
             return False
 
         self.output_formatters.append(scr.OutputFormatter(
             cast(str, self.cm.mc.content_write_format),
-            self.cm, save_file, self.content, self.filename
+            self.cm, save_file, self.content
         ))
         return True
 
@@ -661,8 +792,10 @@ class DownloadJob:
                 self.cm.mc.ctx, self.content, stream=True)
             )
         except ScrFetchError as ex:
-            scr.log(self.cm.mc.ctx, Verbosity.ERROR,
-                    f"{self.context}: failed to open file '{utils.truncate(self.content)}': {str(ex)}")
+            self.log(
+                Verbosity.ERROR,
+                f"{self.context}: failed to open file '{utils.truncate(self.content)}': {str(ex)}"
+            )
             return False
         if self.content_format == ContentFormat.TEMP_FILE:
             self.temp_file_path = self.content
@@ -680,26 +813,113 @@ class DownloadJob:
             stream = sys.stdout.buffer
         self.output_formatters.append(scr.OutputFormatter(
             self.cm.mc.content_print_format, self.cm,
-            stream, self.content, self.filename
+            stream, self.content
         ))
         return True
 
+    def setup_forward_chain(self) -> bool:
+        if self.cm.mc.content_forward_format is None:
+            return True
+        self.forward_buffer = ByteBuffer()
+        self.output_formatters.append(scr.OutputFormatter(
+            self.cm.mc.content_forward_format, self.cm,
+            self.forward_buffer, self.content
+        ))
+        return True
+
+    def process_shell_output(self, stderr: bool) -> None:
+        assert self.shell_proc is not None
+        if stderr:
+            stream = self.shell_proc.stderr
+        else:
+            stream = self.shell_proc.stdout
+        pos = self.shell_cmd_stream
+        assert stream is not None
+        assert pos is not None
+        while True:
+            if self.cm.mc.ctx.abort:
+                break
+            buffer = stream.read(DEFAULT_RESPONSE_BUFFER_SIZE)
+            pos.write(buffer)
+            if len(buffer) < DEFAULT_RESPONSE_BUFFER_SIZE:
+                break
+
+    def setup_shell_cmd_output(self) -> bool:
+        if self.cm.mc.content_shell_command_format is None:
+            return True
+        if self.cm.mc.content_shell_command_stdin_format:
+            stdin = PIPE
+        else:
+            stdin = None
+
+        need_output_handlers = False
+        if self.cm.mc.content_shell_command_print_output:
+            if self.cm.mc.ctx.dl_manager:
+                stdout = PIPE
+                stderr = PIPE
+                need_output_handlers = True
+            else:
+                stdout = None
+                stderr = None
+        else:
+            stdout = subprocess.DEVNULL
+            stderr = subprocess.DEVNULL
+        shell_cmd = scr.gen_final_content_format(self.cm.mc.content_shell_command_format, self.cm)
+        shell_args = shlex.split(shell_cmd.decode("utf-8", errors="surrogateescape"))
+        self.shell_proc = subprocess.Popen(
+            shell_args, stdin=stdin, stdout=stdout, stderr=stderr,
+        )
+        if self.cm.mc.content_shell_command_stdin_format is not None:
+            assert self.shell_proc.stdin is not None
+            self.output_formatters.append(scr.OutputFormatter(
+                self.cm.mc.content_shell_command_stdin_format, self.cm,
+                self.shell_proc.stdin, self.content
+            ))
+        if need_output_handlers:
+            assert self.cm.mc.ctx.dl_manager is not None
+            excecutor = self.cm.mc.ctx.dl_manager.shell_output_handling_executor
+            self.shell_output_handlers.append(excecutor.submit(self.process_shell_output, False))
+            self.shell_output_handlers.append(excecutor.submit(self.process_shell_output, True))
+
+        return True
+
     def check_abort(self) -> None:
         if self.cm.mc.ctx.abort:
             raise InterruptedError
 
-    def run_job(self) -> bool:
-        if self.status_report:
-            self.status_report.gen_display_name(
-                self.cm.url_parsed, self.filename, self.save_path
+    def gen_output_doc(self) -> Optional['document.Document']:
+        if self.forward_buffer is not None:
+            if self.cm.mc.content_raw:
+                path = "<forwarded content match>"
+                path_parsed = None
+            else:
+                path = self.cm.clm.result
+                path_parsed = self.cm.url_parsed
+            doc = document.Document(
+                DocumentType.CONTENT_MATCH,
+                path, self.cm.mc, self.cm.doc,
+                self.cm.mc.content_forward_chains, None, None,
+                path_parsed
             )
-            self.status_report.enqueue()
+            data = self.forward_buffer.to_bytes()
+            doc.encoding = self.cm.mc.default_document_encoding
+            doc.text = data.decode(doc.encoding, errors="surrogateescape")
+            return doc
+        return None
+
+    def run_job(self) -> Optional['document.Document']:
         success = False
         try:
-            if self.handle_save_path() != InteractiveResult.ACCEPT:
-                return False
+            if self.status_report:
+                self.status_report.gen_display_name(
+                    self.cm.url_parsed, self.cm.filename, self.save_path,
+                    self.cm.mc.content_shell_command_format is not None
+                )
+                self.status_report.enqueue()
+            if self.handle_user_interaction() != InteractiveResult.ACCEPT:
+                return None
             if not self.fetch_content():
-                return False
+                return None
 
             self.check_abort()
             self.content_stream: Union[BinaryIO, MinimalInputStream, None] = (
@@ -709,16 +929,21 @@ class DownloadJob:
             )
 
             if not self.setup_content_file():
-                return False
+                return None
             if not self.setup_save_file():
-                return False
+                return None
             if self.status_report:
                 # try to generate a better name now that we have more information
                 self.status_report.gen_display_name(
-                    self.cm.url_parsed, self.filename, self.save_path
+                    self.cm.url_parsed, self.cm.filename, self.save_path,
+                    self.cm.mc.content_shell_command_format is not None
                 )
             if not self.setup_print_output():
-                return False
+                return None
+            if not self.setup_shell_cmd_output():
+                return None
+            if not self.setup_forward_chain():
+                return None
             self.check_abort()
 
             if self.content_stream is None:
@@ -730,7 +955,7 @@ class DownloadJob:
                     assert not res
                     self.check_abort()
                 success = True
-                return True
+                return self.gen_output_doc()
 
             if self.cm.mc.need_output_multipass and self.multipass_file is None:
                 try:
@@ -738,14 +963,13 @@ class DownloadJob:
                         self.cm.mc.ctx, self.save_path)
                     self.temp_file = open(self.temp_file_path, "xb+")
                 except IOError:
-                    return False
+                    return None
                 self.multipass_file = self.temp_file
                 self.check_abort()
 
             if self.content_stream is not None:
                 while True:
-                    buf = self.content_stream.read(
-                        DEFAULT_RESPONSE_BUFFER_SIZE)
+                    buf = self.content_stream.read(DEFAULT_RESPONSE_BUFFER_SIZE)
                     self.check_abort()
                     if buf is None:
                         continue
@@ -771,10 +995,20 @@ class DownloadJob:
                         if len(buf) < DEFAULT_RESPONSE_BUFFER_SIZE:
                             break
             success = True
-            return True
+            return self.gen_output_doc()
         except InterruptedError:
-            return False
+            if self.shell_proc is not None:
+                self.shell_proc.terminate()
+            return None
+        except Exception as ex:
+            if self.shell_proc is not None:
+                self.shell_proc.terminate()
+            raise ex
         finally:
+            if self.shell_proc is not None:
+                if self.shell_proc.stdin is not None:
+                    self.shell_proc.stdin.close()
+                self.shell_proc.wait()
             if self.status_report:
                 self.status_report.finished()
             if self.print_stream is not None:
@@ -789,17 +1023,25 @@ class DownloadJob:
                 self.save_file.close()
             path = self.cm.clm.result
             if self.requires_download():
-                scr.log(self.cm.mc.ctx, Verbosity.DEBUG,
-                        f"finished downloading {path}" if success else f"failed to download {path}"
-                        )
+                self.log(
+                    Verbosity.DEBUG,
+                    f"finished downloading {path}" if success else f"failed to download {path}"
+                )
+            try:
+                for r in concurrent.futures.wait(self.shell_output_handlers).done:
+                    r.result()  # trigger exceptions
+            finally:
+                if self.shell_cmd_stream is not None:
+                    self.shell_cmd_stream.close()
 
 
 class DownloadManager:
     ctx: 'scr_context.ScrContext'
     max_threads: int
-    pending_jobs: set[concurrent.futures.Future[bool]]
+    pending_jobs: set[concurrent.futures.Future[Optional['document.Document']]]
     pom: PrintOutputManager
     executor: concurrent.futures.ThreadPoolExecutor
+    shell_output_handling_executor: concurrent.futures.ThreadPoolExecutor
     status_report_lock: threading.Lock
     download_status_reports: list['progress_report.DownloadStatusReport']
     enable_status_reports: bool
@@ -810,6 +1052,9 @@ class DownloadManager:
         self.executor = concurrent.futures.ThreadPoolExecutor(
             max_workers=max_threads
         )
+        self.shell_output_handling_executor = concurrent.futures.ThreadPoolExecutor(
+            max_workers=2 * max_threads
+        )
         self.pom = PrintOutputManager()
         self.status_report_lock = threading.Lock()
         self.download_status_reports = []
@@ -820,7 +1065,7 @@ class DownloadManager:
             self.ctx, Verbosity.DEBUG,
             f"enqueuing download for {dj.cm.clm.result}"
         )
-        dj.setup_print_stream(self.pom)
+        dj.request_print_streams(self.pom)
         if self.enable_status_reports:
             dj.request_status_report(self)
         self.pending_jobs.add(self.executor.submit(dj.run_job))
@@ -834,7 +1079,7 @@ class DownloadManager:
         if not self.enable_status_reports or not may_print:
             results = concurrent.futures.wait(self.pending_jobs)
             for x in results.done:
-                x.result()
+                scr.forward_document(self.ctx, x.result())
             self.pending_jobs.clear()
         self.pom.request_print_access()
         prm = progress_report.ProgressReportManager()
@@ -845,19 +1090,25 @@ class DownloadManager:
                 else progress_report.DOWNLOAD_STATUS_REFRESH_INTERVAL
             )
             for x in results.done:
-                x.result()
+                scr.forward_document(self.ctx, x.result())
             self.pending_jobs = results.not_done
             prm.load_status(self)
+            if prm.prev_report_line_count == 0 and not self.pending_jobs:
+                # this happens when we got main thread print access
+                # but everybody is already done and never downloaded anything
+                # we don't want any progress reports here
+                break
             if not prm.updates_remaining():
-                if not self.pending_jobs:
-                    # this happens when we got main thread print access
-                    # but everybody is already done and never downloaded anything
-                    # we don't want any progress reports here
-                    break
                 continue
             prm.print_status_report()
             if not self.pending_jobs:
                 break
+        for rl in prm.finished_report_lines + prm.newly_finished_report_lines:
+            if rl.error:
+                scr.log(self.ctx, Verbosity.ERROR, rl.error)
+
+    def reset(self) -> None:
+        self.pom.reset()
 
     def terminate(self, cancel_running: bool = False) -> None:
         try:
