@@ -1,6 +1,6 @@
 from typing import Optional, Union
 from .definitions import (
-    SeleniumStrategy, SeleniumDownloadStrategy, DocumentType,
+    SeleniumStrategy, SeleniumDownloadStrategy, DocumentType, DocumentDuplication,
     DEFAULT_ESCAPE_SEQUENCE, FALLBACK_DOCUMENT_SCHEME)
 from .config_data_class import ConfigDataClass
 from . import locator, content_match, scr_context, document
@@ -14,11 +14,17 @@ class MatchChain(ConfigDataClass):
     cimin: int = 1
     cimax: Union[int, float] = float("inf")
     ci_continuous: bool = False
+    document_duplication: DocumentDuplication = DocumentDuplication.UNIQUE
     content_save_format: Optional[str] = None
     content_print_format: Optional[str] = None
     content_write_format: Optional[str] = None
+    content_shell_command_format: Optional[str] = None
+    content_shell_command_stdin_format: Optional[str] = None
+    content_shell_command_print_output: Optional[bool] = None
+    content_forward_format: Optional[str] = None
     content_forward_chains: list['MatchChain'] = []
     content_raw: bool = True
+
     content_input_encoding: str = "utf-8"
     content_forced_input_encoding: Optional[str] = None
     save_path_interactive: bool = False
@@ -67,16 +73,18 @@ class MatchChain(ConfigDataClass):
     # TODO: this should include if this is the target of any doc=...
     has_document_matching: bool = False
     has_content_matching: bool = False
-    parses_documents = False  # used for the document as content optimization
+    needs_document_content = False  # used for the document as content optimization
     has_interactive_matching: bool = False
     need_content: bool = False
     need_label: bool = False
     need_filename: bool = False
+    need_filename_for_interaction: bool = False
     need_output_multipass: bool = False
     content_matches: list['content_match.ContentMatch']
     document_matches: list['document.Document']
     handled_content_matches: set['content_match.ContentMatch']
     handled_document_matches: set['document.Document']
+    requested_document_urls: set[str]
     satisfied: bool = True
     labels_none_for_n: int = 0
 
@@ -95,16 +103,17 @@ class MatchChain(ConfigDataClass):
         self.document_matches = []
         self.handled_content_matches = set()
         self.handled_document_matches = set()
+        self.requested_document_urls = set()
 
     def gen_dummy_document(self) -> 'document.Document':
         d = document.Document(
-            DocumentType.FILE, "", None,
+            DocumentType.FILE, "",
             locator_match=self.loc_document.gen_dummy_locator_match()
         )
         d.encoding = ""
         return d
 
-    def gen_dummy_content_match(self) -> 'content_match.ContentMatch':
+    def gen_dummy_content_match(self, has_filename: bool) -> 'content_match.ContentMatch':
         clm = self.loc_content.gen_dummy_locator_match()
         if self.has_label_matching:
             llm = self.loc_label.gen_dummy_locator_match()
@@ -120,6 +129,7 @@ class MatchChain(ConfigDataClass):
             dcm.ci = 0
         if self.has_document_matching:
             dcm.di = 0
+        dcm.filename = "" if has_filename else None
         return dcm
 
     def accepts_content_matches(self) -> bool:
