@@ -1,5 +1,6 @@
 #!/usr/bin/env python3
-from typing import Any, Optional, BinaryIO, Union, cast
+from datetime import datetime
+from typing import IO, Any, Optional, BinaryIO, Union, cast
 
 import shutil
 from io import BytesIO
@@ -10,7 +11,6 @@ import lxml.html
 import pyrfc6266
 import requests
 import sys
-import xml.sax.saxutils
 import re
 import os
 from string import Formatter
@@ -21,7 +21,6 @@ import pyparsing.exceptions
 import pathlib
 
 from http.cookiejar import MozillaCookieJar
-
 from selenium.webdriver.remote.webelement import WebElement as SeleniumWebElement
 import selenium.webdriver.common.by
 from selenium.webdriver.remote.webdriver import WebDriver as SeleniumWebDriver
@@ -35,7 +34,7 @@ import warnings
 import urllib.request
 
 from .definitions import (
-    T, K, ScrSetupError, ScrFetchError, ScrMatchError, Verbosity, SCRIPT_NAME,
+    T, K, DocumentDuplication, ScrSetupError, ScrFetchError, ScrMatchError, Verbosity, SCRIPT_NAME,
     SeleniumVariant, SeleniumStrategy, SeleniumDownloadStrategy,
     DocumentType, InteractiveResult,
     verbosities_display_dict, document_type_display_dict,
@@ -49,7 +48,7 @@ from .input_sequences import (
 )
 from . import (
     document, selenium_setup, utils, config_data_class, args_parsing, download_job,
-    locator, content_match, match_chain, scr_context
+    locator, content_match, match_chain, scr_context,
 )
 
 if utils.is_windows():
@@ -65,18 +64,16 @@ class OutputFormatter:
     _args_list: list[Any]
     _format_parts: list[tuple[str, Union[str, None],
                               Union[str, None], Union[str, None]]]
-    _out_stream: Union[BinaryIO, 'download_job.PrintOutputStream']
+    _out_stream: Union['download_job.PrintOutputStream', 'download_job.ByteBuffer', IO[bytes]]
     _found_stream: bool = False
     _input_buffer_sizes: int
 
     def __init__(
         self, format_str: str, cm: 'content_match.ContentMatch',
-        out_stream: Union[BinaryIO, 'download_job.PrintOutputStream'],
+        out_stream: Union['download_job.PrintOutputStream', 'download_job.ByteBuffer', IO[bytes]],
         content: Union[str, bytes, 'download_job.MinimalInputStream', BinaryIO, None],
-        filename: Optional[str]
     ) -> None:
-        self._args_dict = content_match_build_format_args(
-            cm, content, filename)
+        self._args_dict = content_match_build_format_args(cm, content)
         self._args_list = []  # no positional args right now
 
         # we reverse these lists so we can take out elements using pop()
@@ -104,7 +101,7 @@ class OutputFormatter:
                     break
 
             while self._format_parts:
-                (text, key, format_args, b) = self._format_parts.pop()
+                (text, key, format_args, _b) = self._format_parts.pop()
                 if text:
                     self._out_stream.write(text.encode("utf-8"))
                 if key is not None:
@@ -126,7 +123,6 @@ class OutputFormatter:
             if not self._found_stream:
                 break
 
-        assert buffer is None
         assert not self._format_parts
         self._out_stream.flush()
         return False
@@ -185,11 +181,10 @@ def apply_filename_format_args(filename: Optional[str], args_dict: dict[str, Any
 def content_match_build_format_args(
     cm: 'content_match.ContentMatch',
     content: Any = None,
-    filename: Optional[str] = None
 ) -> dict[str, Any]:
     args_dict: dict[str, Any] = {}
     apply_general_format_args(cm.doc, cm.mc, args_dict, ci=cm.ci)
-    apply_filename_format_args(filename, args_dict)
+    apply_filename_format_args(cm.filename, args_dict)
     if content is not None:
         args_dict["c"] = content
 
@@ -210,10 +205,6 @@ def content_match_build_format_args(
     return args_dict
 
 
-def log_raw(verbosity: Verbosity, msg: str) -> None:
-    sys.stderr.write(verbosities_display_dict[verbosity] + msg + "\n")
-
-
 BSE_U_REGEX_MATCH = re.compile("[0-9A-Fa-f]{4}")
 
 
@@ -268,11 +259,23 @@ def unescape_string(txt: str) -> str:
     return txt
 
 
-def log(ctx: 'scr_context.ScrContext', verbosity: Verbosity, msg: str) -> None:
+def check_log_message_needed(ctx: 'scr_context.ScrContext', verbosity: Verbosity) -> bool:
     if verbosity == Verbosity.ERROR:
         ctx.error_code = 1
-    if ctx.verbosity is None or ctx.verbosity >= verbosity:
-        log_raw(verbosity, msg)
+    return ctx.verbosity is None or ctx.verbosity >= verbosity
+
+
+def get_log_str(verbosity: Verbosity, msg: str) -> str:
+    return verbosities_display_dict[verbosity] + msg + "\n"
+
+
+def log_raw(msg: str) -> None:
+    sys.stderr.write(msg)
+
+
+def log(ctx: 'scr_context.ScrContext', verbosity: Verbosity, msg: str) -> None:
+    if check_log_message_needed(ctx, verbosity):
+        log_raw(get_log_str(verbosity, msg))
 
 
 def get_format_string_keys(fmt_string: str) -> list[str]:
@@ -311,11 +314,11 @@ def format_strings_args_occurence(
 def validate_format(
     conf: config_data_class.ConfigDataClass, attrib_path: list[str],
     dummy_cm: 'content_match.ContentMatch',
-    unescape: bool, has_content: bool = False, has_filename: bool = False
+    unescape: bool, has_content: bool = False
 ) -> None:
     try:
         known_keys = content_match_build_format_args(
-            dummy_cm, "" if has_content else None, "" if has_filename else None
+            dummy_cm, "" if has_content else None
         )
         unnamed_key_count = 0
         fmt_keys = get_format_string_keys(conf.resolve_attrib_path(
@@ -331,6 +334,8 @@ def validate_format(
                         f"exceeded number of ordered keys in {conf.get_configuring_argument(attrib_path)}"
                     )
             elif k not in known_keys:
+                # TODO: maybe add some context dependant clues here
+                # example: no {fn} if you forget cl
                 raise ScrSetupError(
                     f"unavailable key '{{{k}}}' in {conf.get_configuring_argument(attrib_path)}"
                 )
@@ -364,6 +369,13 @@ def gen_default_format(mc: 'match_chain.MatchChain') -> str:
     return form
 
 
+def mc_context(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext') -> str:
+    if len(ctx.match_chains) <= 1 or mc.chain_id is None:
+        return ""
+    else:
+        return f"match chain {mc.chain_id}: "
+
+
 def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext') -> None:
 
     mc.apply_defaults(ctx.defaults_mc)
@@ -371,13 +383,16 @@ def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext
     mc.di = mc.dimin
 
     if mc.dimin > mc.dimax:
-        raise ScrSetupError("dimin can't exceed dimax")
+        raise ScrSetupError(f"{mc_context(mc, ctx)}dimin can't exceed dimax")
     if mc.cimin > mc.cimax:
-        raise ScrSetupError("cimin can't exceed cimax")
+        raise ScrSetupError(f"{mc_context(mc, ctx)}cimin can't exceed cimax")
 
     if mc.content_write_format is not None and mc.content_save_format is None:
         mc.content_save_format = DEFAULT_CSF
 
+    if mc.content_forward_chains and mc.content_forward_format is None:
+        mc.content_forward_format = DEFAULT_CWF
+
     if not mc.document_output_chains:
         mc.document_output_chains = [mc]
 
@@ -387,8 +402,8 @@ def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext
     locators = [mc.loc_content, mc.loc_label, mc.loc_document]
     for loc in locators:
         loc.setup(mc)
-        if loc.parses_documents():
-            mc.parses_documents = True
+        if loc.is_active():
+            mc.needs_document_content = True
 
     if any(lc.xpath is not None for lc in locators):
         mc.has_xpath_matching = True
@@ -398,53 +413,96 @@ def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext
         mc.has_content_xpaths = True
     if mc.loc_document.is_active():
         mc.has_document_matching = True
-        mc.parses_documents = True
+        mc.needs_document_content = True
     if mc.loc_label.interactive or mc.loc_content.interactive:
         mc.has_interactive_matching = True
 
+    content_output_variants = [
+        mc.content_print_format,
+        mc.content_save_format,
+        mc.content_shell_command_format,
+        mc.content_forward_format
+    ]
+
     if mc.has_label_matching or mc.loc_content.is_active():
         mc.has_content_matching = True
-    elif mc.content_print_format or mc.content_save_format:
+    elif any(content_output_variants):
         mc.has_content_matching = True
 
-    if mc.has_content_matching and mc.content_print_format is None and mc.content_save_format is None:
+    if mc.has_content_matching and all(cov is None for cov in content_output_variants):
         mc.content_print_format = DEFAULT_CPF
 
-    dummy_cm = mc.gen_dummy_content_match()
-    if mc.content_print_format:
-        validate_format(mc, ["content_print_format"],
-                        dummy_cm, True, True, not mc.content_raw)
+    if not mc.content_raw or ctx.selenium_variant.enabled():
+        mc.needs_document_content = True
+    if not mc.needs_document_content:
+        # prepare chain to be used in the document -> content link optimization
+        mc.content_raw = False
+
+    dummy_cm = mc.gen_dummy_content_match(not mc.content_raw)
+    if not mc.needs_document_content:
+        # to remove "{cm}" from the available matches
+        dummy_cm.clm.result = None  # type: ignore
+    if mc.content_print_format is not None:
+        validate_format(mc, ["content_print_format"], dummy_cm, True, True)
+
+    if mc.content_shell_command_format is not None:
+        validate_format(mc, ["content_shell_command_format"], dummy_cm, True, False)
+        if mc.content_shell_command_stdin_format is not None:
+            validate_format(mc, ["content_shell_command_stdin_format"], dummy_cm, True, True)
+    elif mc.content_shell_command_stdin_format is not None:
+        raise ScrSetupError(
+            f"{mc_context(mc, ctx)}cannot specify cshif without cshf: {mc.get_configuring_argument(['content_shell_command_stdin_format'])}"
+        )
+    if mc.content_forward_format is not None:
+        validate_format(mc, ["content_forward_format"], dummy_cm, True, True)
 
     if mc.content_save_format is not None:
         if mc.content_save_format == "":
             raise ScrSetupError(
-                f"csf cannot be the empty string: {mc.get_configuring_argument(['content_save_format'])}"
+                f"{mc_context(mc, ctx)}csf cannot be the empty string: {mc.get_configuring_argument(['content_save_format'])}"
             )
-        validate_format(mc, ["content_save_format"], dummy_cm,
-                        True, False, not mc.content_raw)
+        validate_format(mc, ["content_save_format"], dummy_cm, True, False)
         if mc.content_write_format is None:
             mc.content_write_format = DEFAULT_CWF
         else:
-            validate_format(mc, ["content_write_format"],
-                            dummy_cm, True, True, not mc.content_raw)
+            validate_format(mc, ["content_write_format"], dummy_cm, True, True)
 
     if not mc.has_label_matching:
         mc.label_allow_missing = True
         if mc.labels_inside_content:
             raise ScrSetupError(
-                f"match chain {mc.chain_id}: cannot specify lic without lx or lr"
+                f"{mc_context(mc, ctx)}cannot specify lic without lx or lr"
             )
     default_format: Optional[str] = None
 
+    if mc.label_default_format is None:
+        if mc.label_allow_missing and mc.need_label:
+            if default_format is None:
+                default_format = gen_default_format(mc)
+            mc.label_default_format = default_format
+    else:
+        validate_format(mc, ["label_default_format"], dummy_cm, True, False)
+
     output_formats = [
         mc.content_print_format,
+        mc.content_shell_command_format,
+        mc.content_shell_command_stdin_format,
         mc.content_save_format,
-        mc.content_write_format  # this is none if save is None
+        mc.content_write_format,
+        mc.content_forward_format
     ]
 
-    mc.need_filename = format_strings_args_occurence(
-        output_formats,
-        ["fn", "fb", "fe"]
+    label_formats = [mc.loc_label.format, mc.label_default_format]
+
+    filename_format_occurences = ["fn", "fb", "fe"]
+
+    # because of lin / csin
+    mc.need_filename_for_interaction = format_strings_args_occurence(
+        [*label_formats, mc.content_save_format], filename_format_occurences
+    ) > 0
+
+    mc.need_filename = mc.need_filename_for_interaction or format_strings_args_occurence(
+        output_formats, filename_format_occurences
     ) > 0
 
     mc.need_content = format_strings_args_occurence(
@@ -456,7 +514,7 @@ def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext
     ) > 0
 
     mc.need_output_multipass = any(
-        format_string_arg_occurence(of, "c") for of in output_formats
+        format_string_arg_occurence(of, "c") > 1 for of in output_formats
     )
 
     if mc.filename_default_format is None:
@@ -464,31 +522,16 @@ def setup_match_chain(mc: 'match_chain.MatchChain', ctx: 'scr_context.ScrContext
             default_format = gen_default_format(mc)
             mc.filename_default_format = default_format + ".dat"
     else:
-        validate_format(
-            mc, ["filename_default_format"],
-            dummy_cm, True, False, False
-        )
+        fn = dummy_cm.filename
+        dummy_cm.filename = None
+        validate_format(mc, ["filename_default_format"], dummy_cm, True, False)
+        dummy_cm.filename = fn
 
-    if mc.label_default_format is None:
-        if mc.label_allow_missing and mc.need_label:
-            if default_format is None:
-                default_format = gen_default_format(mc)
-            mc.label_default_format = default_format
-    else:
-        validate_format(
-            mc, ["label_default_format"],
-            dummy_cm, True, False, False
-        )
     if not mc.has_content_matching and not mc.has_document_matching:
         if not (mc.chain_id == 0 and (mc.ctx.repl or mc.ctx.special_args_occured)):
             raise ScrSetupError(
                 f"match chain {mc.chain_id} is unused, it has neither document nor content matching"
             )
-    if not mc.content_raw:
-        mc.parses_documents = True
-    if not mc.parses_documents:
-        # prepare chain to be used in the document -> content link optimization
-        mc.content_raw = False
 
 
 def load_cookie_jar(ctx: 'scr_context.ScrContext') -> None:
@@ -563,6 +606,9 @@ def setup(ctx: 'scr_context.ScrContext') -> None:
         if d.expand_match_chains_above is not None:
             d.match_chains.extend(
                 ctx.match_chains[d.expand_match_chains_above:])
+        canonical_url = d.canonical_url()
+        for mc in d.match_chains:
+            mc.requested_document_urls.add(canonical_url)
 
     for mc in ctx.match_chains:
         setup_match_chain(mc, ctx)
@@ -685,17 +731,15 @@ def try_read_data_url(cm: 'content_match.ContentMatch') -> Optional[bytes]:
 def request_exception_to_scr_fetch_error(ex: requests.exceptions.RequestException) -> ScrFetchError:
     if isinstance(ex, requests.exceptions.InvalidURL):
         return ScrFetchError("invalid url")
-    if isinstance(ex, requests.exceptions.ConnectionError):
-        return ScrFetchError("connection failed")
-    if isinstance(ex, requests.exceptions.ConnectTimeout):
+    if isinstance(ex, requests.exceptions.ConnectTimeout) or isinstance(ex, requests.ReadTimeout):
         return ScrFetchError("connection timeout")
-    return ScrFetchError(utils.truncate(str(ex)))
+    return ScrFetchError("connection failed")
 
 
 def request_raw(
     ctx: 'scr_context.ScrContext', path: str, path_parsed: urllib.parse.ParseResult,
     cookie_dict: Optional[dict[str, dict[str, dict[str, Any]]]] = None,
-    proxies: Optional[dict[str, Optional[str]]] = None, stream: bool = False
+    proxies: Optional[dict[str, str]] = None, stream: bool = False
 ) -> requests.Response:
     hostname = path_parsed.hostname if path_parsed.hostname else ""
     if cookie_dict is None:
@@ -704,6 +748,7 @@ def request_raw(
         name: ck["value"]
         for name, ck in cookie_dict.get(hostname, {}).items()
     }
+    assert ctx.user_agent is not None
     headers = {'User-Agent': ctx.user_agent}
 
     res = requests.get(
@@ -767,45 +812,129 @@ def requests_dl(
         raise request_exception_to_scr_fetch_error(ex)
 
 
-def selenium_get_full_page_source(ctx: 'scr_context.ScrContext') -> tuple[str, lxml.html.HtmlElement]:
-    drv = cast(SeleniumWebDriver, ctx.selenium_driver)
-    text = drv.page_source
-    doc_xml = cast(lxml.html.HtmlElement, lxml.html.fromstring(text))
-    iframes_xml_all_sources: list[lxml.html.HtmlElement] = doc_xml.xpath(
-        "//iframe"
+def build_xpath_string(text: str) -> str:
+    pos_single_quote = text.find("'")
+    if pos_single_quote == -1:
+        return f"'{text}'"
+    pos_double_quote = text.find('"')
+    if pos_double_quote == -1:
+        return f'"{text}"'
+    # since xpath does not have escaping, if we have both types of
+    # quotes in our string we need to express it using concat()
+    if pos_single_quote < pos_double_quote:
+        head = f'"{text[:pos_double_quote]}"'
+        tail = text[pos_double_quote:]
+    else:
+        head = f"'{text[:pos_single_quote]}'"
+        tail = text[pos_single_quote:]
+    return f'concat({head},{build_xpath_string(tail)})'
+
+
+def build_elem_xpath(root: 'lxml.html.HtmlElement', elem: 'lxml.html.HtmlElement') -> str:
+    id = elem.get("id")
+    if id is not None:
+        return f"//{elem.tag}[@id={build_xpath_string(id)}]"
+    res = ""
+    while True:
+        parent = cast(Optional[lxml.html.HtmlElement], elem.getparent())
+        index = 1
+        if parent is not None:
+            for e in parent.iterchildren():
+                if e == elem:
+                    break
+                if e.tag == elem.tag:
+                    index += 1
+        res = f"/{elem.tag}[{index}]{res}"
+        if elem == root:
+            return res
+        assert parent is not None
+        elem = parent
+
+
+def get_child_iframes(elem: 'lxml.html.HtmlElement') -> list['lxml.html.HtmlElement']:
+    return cast(list[lxml.html.HtmlElement], elem.xpath("//iframe"))
+
+
+def insert_shadow_roots(
+    ctx: 'scr_context.ScrContext', doc_xml: lxml.html.HtmlElement,
+    shadow_root_tree: Any, nested: bool = False
+) -> None:
+    xpath = shadow_root_tree[0]
+    inner_html = shadow_root_tree[1]
+    children = shadow_root_tree[2]
+    if nested:
+        xpath = "/html/" + xpath
+    elem = doc_xml.xpath(xpath)
+    if len(elem) != 1:
+        log(ctx, Verbosity.WARN, "failed to match up shadow roots")
+        return
+    elem = elem[0]
+    shadow_root_xml = cast(
+        lxml.html.HtmlElement, lxml.html.fromstring(inner_html)
     )
-    if not iframes_xml_all_sources:
+    for c in children:
+        insert_shadow_roots(ctx, shadow_root_xml, c, True)
+    elem.append(shadow_root_xml)
+
+
+def expand_shadow_roots(
+    ctx: 'scr_context.ScrContext', doc_xml: lxml.html.HtmlElement
+) -> bool:
+    script = """
+    function build_xpath(el) {
+        const tag_name = el.tagName.toLowerCase();
+        if (el.parentElement === null) return "/" + tag_name;
+        var ci = 0;
+        for (const c of el.parentNode.childNodes) {
+            if (c === el) return build_xpath(el.parentNode) + '/' + tag_name + '[' + (ci + 1) + ']';
+            if (c.nodeType == Node.ELEMENT_NODE && c.tagName.toLowerCase() == tag_name) ci++;
+        }
+    }
+    function collect_shadow_roots(node) {
+        var res = [];
+        for (const e of node.querySelectorAll('*')) {
+            if (e.shadowRoot === null) continue;
+            res.push([build_xpath(e), e.shadowRoot.innerHTML, collect_shadow_roots(e.shadowRoot)]);
+        }
+        return res;
+    }
+    return collect_shadow_roots(document);
+    """
+    shadow_root_tree = selenium_setup.selenium_exec_script(ctx, script)
+    if len(shadow_root_tree) == 0:
+        return False
+    for srt in shadow_root_tree:
+        insert_shadow_roots(ctx, doc_xml, srt)
+    return True
+
+
+def expand_iframes(
+    ctx: 'scr_context.ScrContext', text: str, doc_xml: lxml.html.HtmlElement
+) -> tuple[str, lxml.html.HtmlElement]:
+    drv = cast(SeleniumWebDriver, ctx.selenium_driver)
+    roots_expanded = expand_shadow_roots(ctx, doc_xml)
+    iframes: list[lxml.html.HtmlElement] = get_child_iframes(doc_xml)
+    if not iframes:
+        if roots_expanded:
+            text = cast(str, lxml.html.tostring(doc_xml))
         return text, doc_xml
     depth = 0
+    curr_xml = doc_xml
     try:
         iframe_stack: list[tuple[
             SeleniumWebElement, int, lxml.html.HtmlElement
         ]] = []
         while True:
-            iframes_by_source: dict[str, list[lxml.html.HtmlElement]] = {}
-            for iframe in reversed(iframes_xml_all_sources):
-                iframe_src = iframe.attrib["src"]
-                iframe_src_escaped = xml.sax.saxutils.escape(iframe_src)
-                if iframe_src_escaped in iframes_by_source:
-                    iframes_by_source[iframe_src_escaped].append(iframe)
-                else:
-                    iframes_by_source[iframe_src_escaped] = [iframe]
-            for iframe_src_escaped, iframes_xml in iframes_by_source.items():
+            for iframe in reversed(iframes):
+                iframe_xpath = build_elem_xpath(curr_xml, iframe)
                 iframes_sel = drv.find_elements(
                     by=selenium.webdriver.common.by.By.XPATH,
-                    value=f"//iframe[@src='{iframe_src_escaped}']"
+                    value=iframe_xpath
                 )
-                len_sel = len(iframes_sel)
-                len_xml = len(iframes_xml)
-                if len_sel != len_xml:
-                    log(
-                        ctx, Verbosity.WARN,
-                        f"iframe count diverged for iframe source in '{iframe_src_escaped}'"
-                    )
-                for i in range(0, min(len_sel, len_xml)):
-                    iframe_stack.append(
-                        (iframes_sel[i], depth + 1, iframes_xml[i])
-                    )
+                if len(iframes_sel) != 1:
+                    log(ctx, Verbosity.WARN, "failed to match up iframe contents")
+                else:
+                    iframe_stack.append((iframes_sel[0], depth + 1, iframe))
             if not iframe_stack:
                 break
             iframe_sel, depth_new, curr_xml = iframe_stack.pop()
@@ -813,22 +942,31 @@ def selenium_get_full_page_source(ctx: 'scr_context.ScrContext') -> tuple[str, l
                 depth -= 1
                 drv.switch_to.parent_frame()
             drv.switch_to.frame(iframe_sel)
-            log(ctx, Verbosity.DEBUG,
-                f"expanding iframe {curr_xml.attrib['src']}")
             depth = depth_new
             iframe_xml = cast(
                 lxml.html.HtmlElement, lxml.html.fromstring(drv.page_source)
             )
+            expand_shadow_roots(ctx, iframe_xml)
+            iframes = get_child_iframes(iframe_xml)
             curr_xml.append(iframe_xml)
             curr_xml = iframe_xml
-            lxml.etree.XPath
-            iframes_xml_all_sources = iframe_xml.xpath(".//iframe")
 
         return cast(str, lxml.html.tostring(doc_xml)), doc_xml
+    except SeleniumWebDriverException:
+        # if the document fundamentally changes while we do this, we might
+        # end up trying to focus on a deleted iframe
+        # in that case we just give up on the iframe sources for this iteration
+        return text, doc_xml
     finally:
         drv.switch_to.default_content()
 
 
+def selenium_get_full_page_source(ctx: 'scr_context.ScrContext') -> tuple[str, lxml.html.HtmlElement]:
+    text = cast(SeleniumWebDriver, ctx.selenium_driver).page_source
+    doc_xml = cast(lxml.html.HtmlElement, lxml.html.fromstring(text))
+    return expand_iframes(ctx, text, doc_xml)
+
+
 def fetch_doc(ctx: 'scr_context.ScrContext', doc: 'document.Document') -> None:
     if ctx.selenium_variant.enabled():
         if doc is not ctx.reused_doc or ctx.changed_selenium:
@@ -839,10 +977,12 @@ def fetch_doc(ctx: 'scr_context.ScrContext', doc: 'document.Document') -> None:
             selpath = doc.path
             if doc.document_type in [DocumentType.FILE, DocumentType.RFILE]:
                 selpath = "file:" + os.path.abspath(selpath)
+
+            fetch_begin = datetime.now()
             try:
                 cast(SeleniumWebDriver, ctx.selenium_driver).get(selpath)
             except SeleniumTimeoutException:
-                ScrFetchError("selenium timeout")
+                raise ScrFetchError("selenium timeout")
             except SeleniumWebDriverException as ex:
                 try:
                     if (
@@ -853,6 +993,16 @@ def fetch_doc(ctx: 'scr_context.ScrContext', doc: 'document.Document') -> None:
                 except IOError:
                     pass
                 raise ex
+            while (datetime.now() - fetch_begin).total_seconds() < ctx.request_timeout_seconds:
+                try:
+                    result = selenium_setup.selenium_exec_script(ctx, "return document.readyState == 'complete';")
+                    if result is True:
+                        break
+                    time.sleep(0)
+                except SeleniumTimeoutException:
+                    raise ScrFetchError("selenium timeout")
+                except SeleniumWebDriverException as ex:
+                    raise ex
         log(
             ctx, Verbosity.INFO,
             f"reloading selenium page source for {document_type_display_dict[doc.document_type]} '{doc.path}'"
@@ -868,6 +1018,8 @@ def fetch_doc(ctx: 'scr_context.ScrContext', doc: 'document.Document') -> None:
         ctx.reused_doc = None
         if doc.text and not ctx.changed_selenium:
             return
+    if doc.document_type == DocumentType.CONTENT_MATCH:
+        return
     if doc.document_type in [DocumentType.FILE, DocumentType.RFILE]:
         log(
             ctx, Verbosity.INFO,
@@ -894,9 +1046,9 @@ def fetch_doc(ctx: 'scr_context.ScrContext', doc: 'document.Document') -> None:
     return
 
 
-def gen_final_content_format(format_str: str, cm: 'content_match.ContentMatch', filename: Optional[str] = None) -> bytes:
+def gen_final_content_format(format_str: str, cm: 'content_match.ContentMatch') -> bytes:
     with BytesIO(b"") as buf:
-        of = OutputFormatter(format_str, cm, buf, None, filename)
+        of = OutputFormatter(format_str, cm, buf, None)
         while of.advance():
             pass
         buf.seek(0)
@@ -917,28 +1069,32 @@ def get_ci_di_context(cm: 'content_match.ContentMatch') -> str:
     return di_ci_context
 
 
+def get_content_type_label(cm: 'content_match.ContentMatch') -> str:
+    return "content match" if cm.mc.content_raw else "content link"
+
+
+def forward_document(ctx: 'scr_context.ScrContext', doc: Optional['document.Document']) -> None:
+    if doc is None:
+        return
+    ctx.docs.append(doc)
+
+
 def handle_content_match(cm: 'content_match.ContentMatch', last_doc_path: Optional[str]) -> InteractiveResult:
     cm.di = cm.mc.di
     cm.ci = cm.mc.ci
     cm.mc.loc_content.apply_format_for_content_match(cm, cm.clm)
     cm.mc.ci += 1
 
-    if cm.llm is None:
-        if cm.mc.need_label:
-            cm.llm = locator.LocatorMatch()
-            cm.llm.fres = cast(str, cm.mc.label_default_format).format(
-                **content_match_build_format_args(cm)
-            )
-            cm.llm.result = cm.llm.fres
-    else:
-        cm.mc.loc_label.apply_format_for_content_match(cm, cm.llm)
-
     di_ci_context = get_ci_di_context(cm)
+    content_type = get_content_type_label(cm)
 
-    if cm.llm is not None:
-        label_context = f' (label "{cm.llm.result}")'
+    if cm.llm is not None and cm.llm.rmatch is not None:
+        label_context = f' (label match was "{cm.llm.rmatch}")'
     else:
-        label_context = ""
+        if cm.mc.need_label:
+            label_context = " (no label match)"
+        else:
+            label_context = ""
 
     while True:
         if not cm.mc.content_raw:
@@ -947,7 +1103,7 @@ def handle_content_match(cm: 'content_match.ContentMatch', last_doc_path: Option
                 cm.mc.ctx, cm.mc, cm.doc.document_type, cm.doc.path_parsed,
                 last_doc_path, cm.clm.result, cm.url_parsed
             )
-        content_type = "content match" if cm.mc.content_raw else "content link"
+
         if cm.mc.loc_content.interactive:
             prompt_options = [
                 (InteractiveResult.ACCEPT, YES_INDICATING_STRINGS),
@@ -992,57 +1148,19 @@ def handle_content_match(cm: 'content_match.ContentMatch', last_doc_path: Option
                         cm.clm.result = cm.clm.result[:i]
                         break
         break
-    if cm.mc.loc_label.interactive:
-        assert cm.llm is not None
-        while True:
-            if not cm.mc.is_valid_label(cm.clm.result):
-                log(cm.mc.ctx, Verbosity.WARN,
-                    f'"{cm.doc.path}": labels cannot contain a slash ("{cm.llm.result}")')
-            else:
-                prompt_options = [
-                    (InteractiveResult.ACCEPT, YES_INDICATING_STRINGS),
-                    (InteractiveResult.REJECT, NO_INDICATING_STRINGS),
-                    (InteractiveResult.EDIT, DOC_SKIP_INDICATING_STRINGS),
-                    (InteractiveResult.SKIP_CHAIN, CHAIN_SKIP_INDICATING_STRINGS),
-                    (InteractiveResult.SKIP_DOC, DOC_SKIP_INDICATING_STRINGS)
-                ]
-                if cm.mc.content_raw:
-                    prompt_options.append(
-                        (InteractiveResult.INSPECT, INSPECT_INDICATING_STRINGS))
-                    inspect_opt_str = "/inspect"
-                    prompt_msg = f'"{cm.doc.path}"{di_ci_context}: accept content label "{cm.llm.result}"'
-                else:
-                    inspect_opt_str = ""
-                    prompt_msg = f'"{cm.doc.path}": {content_type} {cm.clm.result}{di_ci_context}: accept content label "{cm.llm.result}"'
-
-                res = prompt(
-                    f'{prompt_msg} [Yes/no/edit/{inspect_opt_str}/chainskip/docskip]? ',
-                    prompt_options,
-                    InteractiveResult.ACCEPT
-                )
-                if res == InteractiveResult.ACCEPT:
-                    break
-                if res == InteractiveResult.INSPECT:
-                    print(
-                        f'"{cm.doc.path}": {content_type} for "{cm.llm.result}":\n' + cm.clm.result)
-                    continue
-                if res != InteractiveResult.EDIT:
-                    return res
-            cm.llm.result = input("enter new label: ")
 
     job = download_job.DownloadJob(cm)
-    if cm.mc.save_path_interactive:
-        res = job.handle_save_path()
-        if res != InteractiveResult.ACCEPT:
-            return res
+    res = job.handle_user_interaction()
+    if res != InteractiveResult.ACCEPT:
+        return res
     if cm.mc.ctx.dl_manager is not None:
         if job.requires_download():
             cm.mc.ctx.dl_manager.submit(job)
         else:
-            job.setup_print_stream(cm.mc.ctx.dl_manager.pom)
-            job.run_job()
+            job.request_print_streams(cm.mc.ctx.dl_manager.pom)
+            forward_document(cm.mc.ctx, job.run_job())
     else:
-        job.run_job()
+        forward_document(cm.mc.ctx, job.run_job())
 
     return InteractiveResult.ACCEPT
 
@@ -1084,7 +1202,7 @@ def gen_content_matches(
             # - we can't just change the document_type because this messes
             #   with last_doc for the repl
             dummy_doc = document.Document(
-                DocumentType.CONTENT_FILE, path=doc.path, src_mc=doc.src_mc,
+                DocumentType.CONTENT_FILE, path=doc.path, src_mc=doc.src_mc, parent_doc=doc,
                 locator_match=doc.locator_match, path_parsed=doc.path_parsed
             )
         return [content_match.ContentMatch(dummy_clm, dummy_llm, mc, dummy_doc)], 0
@@ -1157,21 +1275,43 @@ def gen_document_matches(
     document_lms = mc.loc_document.apply_regex_matches(document_lms)
     document_lms = mc.loc_document.apply_js_matches(doc, mc, document_lms)
     for dlm in document_lms:
-        ndoc = document.Document(
-            doc.document_type.derived_type(),
-            "",
-            mc,
-            mc.document_output_chains,
-            None,
-            dlm
-        )
-        mc.loc_document.apply_format_for_document_match(ndoc, mc, dlm)
-        ndoc.path, ndoc.path_parsed = normalize_link(
+        mc.loc_document.apply_format_for_document_match(doc, mc, dlm)
+        path, path_parsed = normalize_link(
             mc.ctx, mc, doc.document_type, doc.path_parsed,
             last_doc_path, dlm.result,
             urllib.parse.urlparse(dlm.result)
         )
-        document_matches.append(ndoc)
+        if mc.document_duplication != DocumentDuplication.ALLOWED:
+            canonical_url = urllib.parse.urlunparse(path_parsed)
+            if canonical_url in mc.requested_document_urls:
+                if mc.document_duplication == DocumentDuplication.UNIQUE:
+                    continue
+                assert mc.document_duplication == DocumentDuplication.NONRECURSIVE
+                parent = doc
+                recursion_detected = False
+                while True:
+                    pcurl = parent.canonical_url()
+                    if pcurl == canonical_url:
+                        recursion_detected = True
+                        break
+                    if parent.parent_doc is None:
+                        break
+                    parent = parent.parent_doc
+                if recursion_detected:
+                    continue
+            else:
+                mc.requested_document_urls.add(canonical_url)
+
+        document_matches.append(document.Document(
+            doc.document_type.derived_type(),
+            path,
+            mc,
+            doc,
+            mc.document_output_chains,
+            None,
+            dlm,
+            path_parsed
+        ))
 
     return document_matches
 
@@ -1464,7 +1604,7 @@ def process_document_queue(ctx: 'scr_context.ScrContext') -> Optional['document.
         doc_as_content_opt_possible = not ctx.selenium_variant.enabled()
         for mc in doc.match_chains:
             if mc.need_document_matches(False) or mc.need_content_matches():
-                if mc.parses_documents:
+                if mc.needs_document_content:
                     doc_as_content_opt_possible = False
                 unsatisfied_chains += 1
                 mc.satisfied = False
@@ -1498,7 +1638,7 @@ def process_document_queue(ctx: 'scr_context.ScrContext') -> Optional['document.
             try_number += 1
             same_content = static_content and not content_change
             if try_number > 1 and not same_content:
-                assert(ctx.selenium_variant.enabled())
+                assert ctx.selenium_variant.enabled()
                 try:
                     drv = cast(SeleniumWebDriver, ctx.selenium_driver)
                     last_doc_path = drv.current_url
@@ -1628,6 +1768,8 @@ def resolve_repl_defaults(
                 ctx.dl_manager.terminate()
             finally:
                 ctx.dl_manager = None
+    elif ctx.dl_manager is not None:
+        ctx.dl_manager.reset()
     changed_selenium = False
     if ctx_new.selenium_variant != ctx.selenium_variant:
         changed_selenium = True
@@ -1650,15 +1792,12 @@ def resolve_repl_defaults(
                     doctype = DocumentType.FILE
                     if last_doc and last_doc.document_type == DocumentType.RFILE:
                         doctype = DocumentType.RFILE
-                    last_doc = document.Document(
-                        doctype, path, None, None, None
-                    )
+                    last_doc = document.Document(doctype, path)
             else:
                 if not last_doc or doc_url != last_doc.path:
-                    last_doc = document.Document(
-                        DocumentType.URL, doc_url, None, None, None
-                    )
-
+                    last_doc = document.Document(DocumentType.URL, doc_url)
+    for d in ctx_new.docs:
+        d.parent_doc = last_doc
     if not ctx_new.docs and last_doc:
         last_doc.expand_match_chains_above = len(ctx_new.match_chains)
         last_doc.match_chains = list(ctx_new.match_chains)
@@ -1731,6 +1870,7 @@ def run_repl(initial_ctx: 'scr_context.ScrContext', args: list[str]) -> int:
                     ctx = None
             except KeyboardInterrupt:
                 print("")
+                ctx = None
                 continue
     finally:
         if not success:
@@ -1741,17 +1881,17 @@ def run_repl(initial_ctx: 'scr_context.ScrContext', args: list[str]) -> int:
 def run_scr(args: list[str]) -> int:
     ctx = scr_context.ScrContext(blank=True)
     if len(args) < 2:
-        log_raw(
+        sys.stderr.write(get_log_str(
             Verbosity.ERROR,
             f"missing command line options. Consider {SCRIPT_NAME} --help"
-        )
+        ))
         return 1
 
     try:
         args_parsing.parse_args(ctx, args[1:])
         setup(ctx)
     except ScrSetupError as ex:
-        log_raw(Verbosity.ERROR, str(ex))
+        sys.stderr.write(get_log_str(Verbosity.ERROR, str(ex)))
         return 1
     if ctx.repl:
         ec = run_repl(ctx, args)
